# -*- coding: utf-8 -*-
# flake8: noqa
# type: ignore

"""Command line interface for PyStow."""

import os
from typing import Optional, Sequence

import click


@click.group()
def main() -> None:
    """Run the PyStow CLI."""


@main.command()
@click.argument("keys", nargs=-1)
@click.option("--name")
def join(keys: Sequence[str], name: Optional[str]):
    """List a directory."""
    from . import api

    click.echo(api.join(*keys, name=name))


@main.command()
@click.argument("keys", nargs=-1)
def ls(keys: Sequence[str]):
    """List a directory."""
    from . import api

    directory = api.join(*keys)
    _ls(directory)


@main.command()
@click.argument("keys", nargs=-1)
@click.option("--url", required=True)
@click.option("--name")
@click.option("--force", is_flag=True)
def ensure(keys: Sequence[str], url: str, name: Optional[str], force: bool):
    """Ensure a file is downloaded."""
    from . import api

    path = api.ensure(*keys, url=url, name=name, force=force)
    _ls(path.parent)


def _ls(directory):
    command = f"ls -al {directory}"
    click.secho(f"[pystow] {command}", fg="cyan", bold=True)
    os.system(command)  # noqa:S605


@main.command(name="set")
@click.argument("module")
@click.argument("key")
@click.argument("value")
def set_config(module: str, key: str, value: str):
    """Set a configuration value."""
    from .config_api import write_config

    write_config(module, key, value)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""API functions for PyStow."""

import sqlite3
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Mapping, Optional, Sequence, Union

from .constants import JSON, BytesOpener, Opener, Provider
from .impl import Module

if TYPE_CHECKING:
    import lxml.etree
    import numpy.typing
    import pandas as pd
    import rdflib

__all__ = [
    "module",
    "join",
    "joinpath_sqlite",
    # Opener functions
    "open",
    "open_gz",
    # Loader functions
    "load_df",
    "load_json",
    "load_pickle",
    "load_pickle_gz",
    "load_rdf",
    "load_xml",
    # Dump functions
    "dump_df",
    "dump_json",
    "dump_pickle",
    "dump_rdf",
    "dump_xml",
    # Downloader functions
    "ensure",
    "ensure_from_s3",
    "ensure_from_google",
    # Downloader functions with postprocessing
    "ensure_untar",
    "ensure_gunzip",
    # Downloader + opener functions
    "ensure_open",
    "ensure_open_gz",
    "ensure_open_bz2",
    "ensure_open_lzma",
    "ensure_open_tarfile",
    "ensure_open_zip",
    "ensure_open_sqlite",
    "ensure_open_sqlite_gz",
    # Processors
    "ensure_csv",
    "ensure_custom",
    "ensure_json",
    "ensure_json_bz2",
    "ensure_pickle",
    "ensure_pickle_gz",
    "ensure_excel",
    "ensure_xml",
    "ensure_rdf",
    "ensure_tar_df",
    "ensure_tar_xml",
    "ensure_zip_df",
    "ensure_zip_np",
]


def module(key: str, *subkeys: str, ensure_exists: bool = True) -> Module:
    """Return a module for the application.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param ensure_exists:
        Should all directories be created automatically?
        Defaults to true.
    :return:
        The module object that manages getting and ensuring
    """
    return Module.from_key(key, *subkeys, ensure_exists=ensure_exists)


def join(key: str, *subkeys: str, name: Optional[str] = None, ensure_exists: bool = True) -> Path:
    """Return the home data directory for the given module.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join
    :param name:
        The name of the file (optional) inside the folder
    :param ensure_exists:
        Should all directories be created automatically?
        Defaults to true.
    :return:
        The path of the directory or subdirectory for the given module.
    """
    _module = Module.from_key(key, ensure_exists=ensure_exists)
    return _module.join(*subkeys, name=name, ensure_exists=ensure_exists)


@contextmanager
def open(
    key: str,
    *subkeys: str,
    name: str,
    mode: str = "r",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> Opener:
    """Open a file that exists already.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param mode: The read mode, passed to :func:`open`
    :param open_kwargs: Additional keyword arguments passed to :func:`open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.open(*subkeys, name=name, mode=mode, open_kwargs=open_kwargs) as file:
        yield file


@contextmanager
def open_gz(
    key: str,
    *subkeys: str,
    name: str,
    mode: str = "rt",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> Opener:
    """Open a gzipped file that exists already.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param mode: The read mode, passed to :func:`gzip.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.open_gz(*subkeys, name=name, mode=mode, open_kwargs=open_kwargs) as file:
        yield file


def ensure(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
) -> Path:
    """Ensure a file is downloaded.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :return:
        The path of the file that has been downloaded (or already exists)
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure(
        *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
    )


def ensure_custom(
    key: str,
    *subkeys: str,
    name: str,
    force: bool = False,
    provider: Provider,
    **kwargs: Any,
) -> Path:
    """Ensure a file is present, and run a custom create function otherwise.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name:
        The file name.
    :param force:
        Should the file be re-created, even if the path already exists?
    :param provider:
        The file provider. Will be run with the path as the first positional argument,
        if the file needs to be generated.
    :param kwargs:
        Additional keyword-based parameters passed to the provider.

    :return:
        The path of the file that has been created (or already exists)
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_custom(*subkeys, name=name, force=force, provider=provider, **kwargs)


def ensure_untar(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    directory: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    extract_kwargs: Optional[Mapping[str, Any]] = None,
) -> Path:
    """Ensure a file is downloaded and untarred.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param directory:
        Overrides the name of the directory into which the tar archive is extracted.
        If none given, will use the stem of the file name that gets downloaded.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param extract_kwargs: Keyword arguments to pass to :meth:`tarfile.TarFile.extract_all`.
    :return:
        The path of the directory where the file that has been downloaded
        gets extracted to
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_untar(
        *subkeys,
        url=url,
        name=name,
        directory=directory,
        force=force,
        download_kwargs=download_kwargs,
        extract_kwargs=extract_kwargs,
    )


def ensure_gunzip(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    autoclean: bool = True,
    download_kwargs: Optional[Mapping[str, Any]] = None,
) -> Path:
    """Ensure a file is downloaded and gunzipped.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param autoclean: Should the zipped file be deleted?
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :return:
        The path of the directory where the file that has been downloaded
        gets extracted to
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_gunzip(
        *subkeys,
        url=url,
        name=name,
        force=force,
        autoclean=autoclean,
        download_kwargs=download_kwargs,
    )


@contextmanager
def ensure_open(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "r",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> Opener:
    """Ensure a file is downloaded and open it.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`lzma.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`lzma.open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
    ) as yv:
        yield yv


@contextmanager
def ensure_open_zip(
    key: str,
    *subkeys: str,
    url: str,
    inner_path: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "r",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> BytesOpener:
    """Ensure a file is downloaded then open it with :mod:`zipfile`.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param inner_path:
        The relative path to the file inside the archive
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`zipfile.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`zipfile.open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open_zip(
        *subkeys,
        url=url,
        inner_path=inner_path,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
    ) as yv:
        yield yv


@contextmanager
def ensure_open_lzma(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "r",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> Opener:
    """Ensure a LZMA-compressed file is downloaded and open a file inside it.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`lzma.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`lzma.open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open_lzma(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
    ) as yv:
        yield yv


@contextmanager
def ensure_open_tarfile(
    key: str,
    *subkeys: str,
    url: str,
    inner_path: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "r",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> BytesOpener:
    """Ensure a tar file is downloaded and open a file inside it.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param inner_path:
        The relative path to the file inside the archive
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`tarfile.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`tarfile.open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open_tarfile(
        *subkeys,
        url=url,
        inner_path=inner_path,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
    ) as yv:
        yield yv


@contextmanager
def ensure_open_gz(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "rb",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> Opener:
    """Ensure a gzipped file is downloaded and open a file inside it.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`gzip.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open_gz(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
    ) as yv:
        yield yv


@contextmanager
def ensure_open_bz2(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "rb",
    open_kwargs: Optional[Mapping[str, Any]] = None,
) -> Opener:
    """Ensure a BZ2-compressed file is downloaded and open a file inside it.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`bz2.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`bz2.open`

    :yields: An open file object
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open_bz2(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
    ) as yv:
        yield yv


def ensure_csv(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    read_csv_kwargs: Optional[Mapping[str, Any]] = None,
) -> "pd.DataFrame":
    """Download a CSV and open as a dataframe with :mod:`pandas`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.

        .. note::

            It is assumed that the CSV uses tab separators, as this is the
            only safe option. For more information, see
            `Wikipedia <https://en.wikipedia.org/wiki/Comma-separated_values>`_ and
            `Issue #51 <https://github.com/cthoyt/pystow/issues/51>`_. To override this
            behavior and load using the comma separator, specify ``read_csv_kwargs=dict(sep=",")``.

    :return: A pandas DataFrame

    Example usage::

    >>> import pystow
    >>> import pandas as pd
    >>> url = 'https://raw.githubusercontent.com/pykeen/pykeen/master/src/pykeen/datasets/nations/test.txt'
    >>> df: pd.DataFrame = pystow.ensure_csv('pykeen', 'datasets', 'nations', url=url)
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_csv(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        read_csv_kwargs=read_csv_kwargs,
    )


def load_df(
    key: str,
    *subkeys: str,
    name: str,
    read_csv_kwargs: Optional[Mapping[str, Any]] = None,
) -> "pd.DataFrame":
    """Open a pre-existing CSV as a dataframe with :mod:`pandas`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
    :return: A pandas DataFrame

    Example usage::

    >>> import pystow
    >>> import pandas as pd
    >>> url = 'https://raw.githubusercontent.com/pykeen/pykeen/master/src/pykeen/datasets/nations/test.txt'
    >>> pystow.ensure_csv('pykeen', 'datasets', 'nations', url=url)
    >>> df: pd.DataFrame = pystow.load_df('pykeen', 'datasets', 'nations', name='test.txt')
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.load_df(
        *subkeys,
        name=name,
        read_csv_kwargs=read_csv_kwargs,
    )


def dump_df(
    key: str,
    *subkeys: str,
    name: str,
    obj: "pd.DataFrame",
    sep: str = "\t",
    index: bool = False,
    to_csv_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """Dump a dataframe to a TSV file with :mod:`pandas`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param obj: The dataframe to dump
    :param sep: The separator to use, defaults to a tab
    :param index: Should the index be dumped? Defaults to false.
    :param to_csv_kwargs: Keyword arguments to pass through to :meth:`pandas.DataFrame.to_csv`.
    """
    _module = Module.from_key(key, ensure_exists=True)
    _module.dump_df(
        *subkeys,
        name=name,
        obj=obj,
        sep=sep,
        index=index,
        to_csv_kwargs=to_csv_kwargs,
    )


def ensure_json(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    open_kwargs: Optional[Mapping[str, Any]] = None,
    json_load_kwargs: Optional[Mapping[str, Any]] = None,
) -> JSON:
    """Download JSON and open with :mod:`json`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param open_kwargs: Additional keyword arguments passed to :func:`open`
    :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.
    :returns: A JSON object (list, dict, etc.)

    Example usage::

    >>> import pystow
    >>> url = 'https://maayanlab.cloud/CREEDS/download/single_gene_perturbations-v1.0.json'
    >>> perturbations = pystow.ensure_json('bio', 'creeds', '1.0', url=url)
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_json(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        open_kwargs=open_kwargs,
        json_load_kwargs=json_load_kwargs,
    )


def ensure_json_bz2(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    open_kwargs: Optional[Mapping[str, Any]] = None,
    json_load_kwargs: Optional[Mapping[str, Any]] = None,
) -> JSON:
    """Download BZ2-compressed JSON and open with :mod:`json`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param open_kwargs: Additional keyword arguments passed to :func:`bz2.open`
    :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.
    :returns: A JSON object (list, dict, etc.)

    Example usage::

    >>> import pystow
    >>> url = 'https://github.com/hetio/hetionet/raw/master/hetnet/json/hetionet-v1.0.json.bz2'
    >>> hetionet = pystow.ensure_json_bz2('bio', 'hetionet', '1.0', url=url)
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_json_bz2(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        open_kwargs=open_kwargs,
        json_load_kwargs=json_load_kwargs,
    )


def load_json(
    key: str,
    *subkeys: str,
    name: str,
    json_load_kwargs: Optional[Mapping[str, Any]] = None,
) -> JSON:
    """Open a JSON file :mod:`json`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.
    :returns: A JSON object (list, dict, etc.)
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.load_json(*subkeys, name=name, json_load_kwargs=json_load_kwargs)


def dump_json(
    key: str,
    *subkeys: str,
    name: str,
    obj: JSON,
    open_kwargs: Optional[Mapping[str, Any]] = None,
    json_dump_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """Dump an object to a file with :mod:`json`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param obj: The object to dump
    :param open_kwargs: Additional keyword arguments passed to :func:`open`
    :param json_dump_kwargs: Keyword arguments to pass through to :func:`json.dump`.
    """
    _module = Module.from_key(key, ensure_exists=True)
    _module.dump_json(
        *subkeys, name=name, obj=obj, open_kwargs=open_kwargs, json_dump_kwargs=json_dump_kwargs
    )


def ensure_pickle(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "rb",
    open_kwargs: Optional[Mapping[str, Any]] = None,
    pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
) -> Any:
    """Download a pickle file and open with :mod:`pickle`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`open`
    :param open_kwargs: Additional keyword arguments passed to :func:`open`
    :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
    :returns: Any object
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_pickle(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
        pickle_load_kwargs=pickle_load_kwargs,
    )


def load_pickle(
    key: str,
    *subkeys: str,
    name: str,
    mode: str = "rb",
    open_kwargs: Optional[Mapping[str, Any]] = None,
    pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
) -> Any:
    """Open a pickle file with :mod:`pickle`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param mode: The read mode, passed to :func:`open`
    :param open_kwargs: Additional keyword arguments passed to :func:`open`
    :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
    :returns: Any object
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.load_pickle(
        *subkeys,
        name=name,
        mode=mode,
        open_kwargs=open_kwargs,
        pickle_load_kwargs=pickle_load_kwargs,
    )


def dump_pickle(
    key: str,
    *subkeys: str,
    name: str,
    obj: Any,
    mode: str = "wb",
    open_kwargs: Optional[Mapping[str, Any]] = None,
    pickle_dump_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """Dump an object to a file with :mod:`pickle`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param obj: The object to dump
    :param mode: The read mode, passed to :func:`open`
    :param open_kwargs: Additional keyword arguments passed to :func:`open`
    :param pickle_dump_kwargs: Keyword arguments to pass through to :func:`pickle.dump`.
    """
    _module = Module.from_key(key, ensure_exists=True)
    _module.dump_pickle(
        *subkeys,
        name=name,
        obj=obj,
        mode=mode,
        open_kwargs=open_kwargs,
        pickle_dump_kwargs=pickle_dump_kwargs,
    )


def ensure_pickle_gz(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    mode: str = "rb",
    open_kwargs: Optional[Mapping[str, Any]] = None,
    pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
) -> Any:
    """Download a gzipped pickle file and open with :mod:`pickle`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param mode: The read mode, passed to :func:`gzip.open`
    :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
    :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
    :returns: Any object
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_pickle_gz(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        mode=mode,
        open_kwargs=open_kwargs,
        pickle_load_kwargs=pickle_load_kwargs,
    )


def load_pickle_gz(
    key: str,
    *subkeys: str,
    name: str,
    mode: str = "rb",
    open_kwargs: Optional[Mapping[str, Any]] = None,
    pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
) -> Any:
    """Open a gzipped pickle file with :mod:`pickle`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param mode: The read mode, passed to :func:`open`
    :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
    :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
    :returns: Any object
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.load_pickle_gz(
        *subkeys,
        name=name,
        mode=mode,
        open_kwargs=open_kwargs,
        pickle_load_kwargs=pickle_load_kwargs,
    )


def ensure_xml(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    parse_kwargs: Optional[Mapping[str, Any]] = None,
) -> "lxml.etree.ElementTree":
    """Download an XML file and open it with :mod:`lxml`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param parse_kwargs: Keyword arguments to pass through to :func:`lxml.etree.parse`.
    :returns: An ElementTree object

    .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_xml(
        *subkeys,
        name=name,
        url=url,
        force=force,
        download_kwargs=download_kwargs,
        parse_kwargs=parse_kwargs,
    )


def load_xml(
    key: str,
    *subkeys: str,
    name: str,
    parse_kwargs: Optional[Mapping[str, Any]] = None,
) -> "lxml.etree.ElementTree":
    """Load an XML file with :mod:`lxml`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param parse_kwargs: Keyword arguments to pass through to :func:`lxml.etree.parse`.
    :returns: An ElementTree object

    .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.load_xml(
        *subkeys,
        name=name,
        parse_kwargs=parse_kwargs,
    )


def dump_xml(
    key: str,
    *subkeys: str,
    name: str,
    obj: "lxml.etree.ElementTree",
    open_kwargs: Optional[Mapping[str, Any]] = None,
    write_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """Dump an XML element tree to a file with :mod:`lxml`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param obj: The object to dump
    :param open_kwargs: Additional keyword arguments passed to :func:`open`
    :param write_kwargs: Keyword arguments to pass through to :func:`lxml.etree.ElementTree.write`.
    """
    _module = Module.from_key(key, ensure_exists=True)
    _module.dump_xml(
        *subkeys,
        name=name,
        obj=obj,
        open_kwargs=open_kwargs,
        write_kwargs=write_kwargs,
    )


def ensure_excel(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    read_excel_kwargs: Optional[Mapping[str, Any]] = None,
) -> "pd.DataFrame":
    """Download an excel file and open as a dataframe with :mod:`pandas`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param read_excel_kwargs: Keyword arguments to pass through to :func:`pandas.read_excel`.
    :return: A pandas DataFrame
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_excel(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        read_excel_kwargs=read_excel_kwargs,
    )


def ensure_tar_df(
    key: str,
    *subkeys: str,
    url: str,
    inner_path: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    read_csv_kwargs: Optional[Mapping[str, Any]] = None,
) -> "pd.DataFrame":
    """Download a tar file and open an inner file as a dataframe with :mod:`pandas`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param inner_path:
        The relative path to the file inside the archive
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
    :returns: A dataframe

    .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_tar_df(
        *subkeys,
        url=url,
        name=name,
        force=force,
        inner_path=inner_path,
        download_kwargs=download_kwargs,
        read_csv_kwargs=read_csv_kwargs,
    )


def ensure_tar_xml(
    key: str,
    *subkeys: str,
    url: str,
    inner_path: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    parse_kwargs: Optional[Mapping[str, Any]] = None,
) -> "lxml.etree.ElementTree":
    """Download a tar file and open an inner file as an XML with :mod:`lxml`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param inner_path:
        The relative path to the file inside the archive
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param parse_kwargs: Keyword arguments to pass through to :func:`lxml.etree.parse`.
    :returns: An ElementTree object

    .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_tar_xml(
        *subkeys,
        url=url,
        name=name,
        force=force,
        inner_path=inner_path,
        download_kwargs=download_kwargs,
        parse_kwargs=parse_kwargs,
    )


def ensure_zip_df(
    key: str,
    *subkeys: str,
    url: str,
    inner_path: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    read_csv_kwargs: Optional[Mapping[str, Any]] = None,
) -> "pd.DataFrame":
    """Download a zip file and open an inner file as a dataframe with :mod:`pandas`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param inner_path:
        The relative path to the file inside the archive
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
    :return: A pandas DataFrame
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_zip_df(
        *subkeys,
        url=url,
        name=name,
        force=force,
        inner_path=inner_path,
        download_kwargs=download_kwargs,
        read_csv_kwargs=read_csv_kwargs,
    )


def ensure_zip_np(
    key: str,
    *subkeys: str,
    url: str,
    inner_path: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    load_kwargs: Optional[Mapping[str, Any]] = None,
) -> "numpy.typing.ArrayLike":
    """Download a zip file and open an inner file as an array-like with :mod:`numpy`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param inner_path:
        The relative path to the file inside the archive
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs:
        Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param load_kwargs:
        Additional keyword arguments that are passed through to :func:`read_zip_np`
        and transitively to :func:`numpy.load`.
    :returns: An array-like object
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_zip_np(
        *subkeys,
        url=url,
        name=name,
        force=force,
        inner_path=inner_path,
        download_kwargs=download_kwargs,
        load_kwargs=load_kwargs,
    )


def ensure_rdf(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
    precache: bool = True,
    parse_kwargs: Optional[Mapping[str, Any]] = None,
) -> "rdflib.Graph":
    """Download a RDF file and open with :mod:`rdflib`.

    :param key: The module name
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :param precache: Should the parsed :class:`rdflib.Graph` be stored as a pickle for fast loading?
    :param parse_kwargs:
        Keyword arguments to pass through to :func:`pystow.utils.read_rdf` and transitively to
        :func:`rdflib.Graph.parse`.
    :return: An RDF graph

    Example usage::

    >>> import pystow
    >>> import rdflib
    >>> url = 'https://ftp.expasy.org/databases/rhea/rdf/rhea.rdf.gz'
    >>> rdf_graph: rdflib.Graph = pystow.ensure_rdf('rhea', url=url)

    If :mod:`rdflib` fails to guess the format, you can explicitly specify it using the `parse_kwargs` argument:

    >>> import pystow
    >>> import rdflib
    >>> url = "http://oaei.webdatacommons.org/tdrs/testdata/persistent/knowledgegraph" \
    ... "/v3/suite/memoryalpha-stexpanded/component/reference.xml"
    >>> rdf_graph: rdflib.Graph = pystow.ensure_rdf("memoryalpha-stexpanded", url=url, parse_kwargs={"format": "xml"})
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_rdf(
        *subkeys,
        url=url,
        name=name,
        force=force,
        download_kwargs=download_kwargs,
        precache=precache,
        parse_kwargs=parse_kwargs,
    )


def load_rdf(
    key: str,
    *subkeys: str,
    name: Optional[str] = None,
    parse_kwargs: Optional[Mapping[str, Any]] = None,
) -> "rdflib.Graph":
    """Open an RDF file with :mod:`rdflib`.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param parse_kwargs:
        Keyword arguments to pass through to :func:`pystow.utils.read_rdf` and transitively to
        :func:`rdflib.Graph.parse`.
    :return: An RDF graph
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.load_rdf(*subkeys, name=name, parse_kwargs=parse_kwargs)


def dump_rdf(
    key: str,
    *subkeys: str,
    name: str,
    obj: "rdflib.Graph",
    format: str = "turtle",
    serialize_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """Dump an RDF graph to a file with :mod:`rdflib`.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the file to open
    :param obj: The object to dump
    :param format: The format to dump in
    :param serialize_kwargs:
        Keyword arguments to through to :func:`rdflib.Graph.serialize`.
    """
    _module = Module.from_key(key, ensure_exists=True)
    _module.dump_rdf(*subkeys, name=name, obj=obj, format=format, serialize_kwargs=serialize_kwargs)


def ensure_from_s3(
    key: str,
    *subkeys: str,
    s3_bucket: str,
    s3_key: Union[str, Sequence[str]],
    name: Optional[str] = None,
    force: bool = False,
    **kwargs: Any,
) -> Path:
    """Ensure a file is downloaded.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param s3_bucket:
        The S3 bucket name
    :param s3_key:
        The S3 key name
    :param name:
        Overrides the name of the file at the end of the S3 key, if given.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param kwargs:
        Remaining kwargs to forwrad to :class:`Module.ensure_from_s3`.
    :return:
        The path of the file that has been downloaded (or already exists)

    Example downloading ProtMapper 0.0.21:

    >>> version = '0.0.21'
    >>> ensure_from_s3('test', version, s3_bucket='bigmech', s3_key=f'protmapper/{version}/refseq_uniprot.csv')
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_from_s3(
        *subkeys, s3_bucket=s3_bucket, s3_key=s3_key, name=name, force=force, **kwargs
    )


def ensure_from_google(
    key: str,
    *subkeys: str,
    name: str,
    file_id: str,
    force: bool = False,
) -> Path:
    """Ensure a file is downloaded from google drive.

    :param key:
        The name of the module. No funny characters. The envvar
        <key>_HOME where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name:
        The name of the file
    :param file_id:
        The file identifier of the google file. If your share link is
        https://drive.google.com/file/d/1AsPPU4ka1Rc9u-XYMGWtvV65hF3egi0z/view, then your file id is
        ``1AsPPU4ka1Rc9u-XYMGWtvV65hF3egi0z``.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :return:
        The path of the file that has been downloaded (or already exists)

    Example downloading the WK3l-15k dataset as motivated by
    https://github.com/pykeen/pykeen/pull/403:

    >>> ensure_from_google('test', name='wk3l15k.zip', file_id='1AsPPU4ka1Rc9u-XYMGWtvV65hF3egi0z')
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.ensure_from_google(*subkeys, name=name, file_id=file_id, force=force)


def joinpath_sqlite(key: str, *subkeys: str, name: str) -> str:
    """Get an SQLite database connection string.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param name: The name of the database file.
    :return: A SQLite path string.
    """
    _module = Module.from_key(key, ensure_exists=True)
    return _module.joinpath_sqlite(*subkeys, name=name)


@contextmanager
def ensure_open_sqlite(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
) -> Generator[sqlite3.Connection, None, None]:
    """Ensure and connect to a SQLite database.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :yields: An instance of :class:`sqlite3.Connection` from :func:`sqlite3.connect`

    Example usage:
    >>> import pystow
    >>> import pandas as pd
    >>> url = "https://s3.amazonaws.com/bbop-sqlite/hp.db"
    >>> sql = "SELECT * FROM entailed_edge LIMIT 10"
    >>> with pystow.ensure_open_sqlite("test", url=url) as conn:
    >>>     df = pd.read_sql(sql, conn)
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open_sqlite(
        *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
    ) as yv:
        yield yv


@contextmanager
def ensure_open_sqlite_gz(
    key: str,
    *subkeys: str,
    url: str,
    name: Optional[str] = None,
    force: bool = False,
    download_kwargs: Optional[Mapping[str, Any]] = None,
) -> Generator[sqlite3.Connection, None, None]:
    """Ensure and connect to a gzipped SQLite database.

    :param key:
        The name of the module. No funny characters. The envvar
        `<key>_HOME` where key is uppercased is checked first before using
        the default home directory.
    :param subkeys:
        A sequence of additional strings to join. If none are given,
        returns the directory for this module.
    :param url:
        The URL to download.
    :param name:
        Overrides the name of the file at the end of the URL, if given. Also
        useful for URLs that don't have proper filenames with extensions.
    :param force:
        Should the download be done again, even if the path already exists?
        Defaults to false.
    :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
    :yields: An instance of :class:`sqlite3.Connection` from :func:`sqlite3.connect`

    Example usage:
    >>> import pystow
    >>> import pandas as pd
    >>> url = "https://s3.amazonaws.com/bbop-sqlite/hp.db.gz"
    >>> sql = "SELECT * FROM entailed_edge LIMIT 10"
    >>> with pystow.ensure_open_sqlite_gz("test", url=url) as conn:
    >>>     df = pd.read_sql(sql, conn)
    """
    _module = Module.from_key(key, ensure_exists=True)
    with _module.ensure_open_sqlite_gz(
        *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
    ) as yv:
        yield yv



# -*- coding: utf-8 -*-

"""Tests for PyStow."""

import bz2
import contextlib
import itertools as itt
import json
import lzma
import os
import pickle
import shutil
import tempfile
import unittest
from pathlib import Path
from typing import ContextManager, Mapping, Union
from unittest import mock

import pandas as pd

import pystow
from pystow import join
from pystow.constants import PYSTOW_HOME_ENVVAR, PYSTOW_NAME_ENVVAR
from pystow.impl import Module
from pystow.utils import (
    get_home,
    get_name,
    mock_envvar,
    n,
    write_pickle_gz,
    write_sql,
    write_tarfile_csv,
    write_zipfile_csv,
)

HERE = Path(__file__).parent.resolve()
RESOURCES = HERE.joinpath("resources")

TSV_NAME = "test_1.tsv"
TSV_URL = f"{n()}/{TSV_NAME}"

SQLITE_NAME = "test_1.db"
SQLITE_URL = f"{n()}/{SQLITE_NAME}"
SQLITE_PATH = RESOURCES / SQLITE_NAME
SQLITE_TABLE = "testtable"

JSON_NAME = "test_1.json"
JSON_URL = f"{n()}/{JSON_NAME}"
JSON_PATH = RESOURCES / JSON_NAME

PICKLE_NAME = "test_1.pkl"
PICKLE_URL = f"{n()}/{PICKLE_NAME}"
PICKLE_PATH = RESOURCES / PICKLE_NAME

PICKLE_GZ_NAME = "test_1.pkl.gz"
PICKLE_GZ_URL = f"{n()}/{PICKLE_GZ_NAME}"
PICKLE_GZ_PATH = RESOURCES / PICKLE_GZ_NAME

JSON_BZ2_NAME = "test_1.json.bz2"
JSON_BZ2_URL = f"{n()}/{JSON_BZ2_NAME}"
JSON_BZ2_PATH = RESOURCES / JSON_BZ2_NAME


MOCK_FILES: Mapping[str, Path] = {
    TSV_URL: RESOURCES / TSV_NAME,
    JSON_URL: JSON_PATH,
    JSON_BZ2_URL: JSON_BZ2_PATH,
    PICKLE_URL: PICKLE_PATH,
    PICKLE_GZ_URL: PICKLE_GZ_PATH,
    SQLITE_URL: SQLITE_PATH,
}

TEST_TSV_ROWS = [
    ("h1", "h2", "h3"),
    ("v1_1", "v1_2", "v1_3"),
    ("v2_1", "v2_2", "v2_3"),
]
TEST_DF = pd.DataFrame(TEST_TSV_ROWS)
TEST_JSON = {"key": "value"}

# Make the pickle file
if not PICKLE_PATH.is_file():
    PICKLE_PATH.write_bytes(pickle.dumps(TEST_TSV_ROWS))

if not SQLITE_PATH.is_file():
    write_sql(TEST_DF, name=SQLITE_TABLE, path=SQLITE_PATH, index=False)

if not JSON_PATH.is_file():
    JSON_PATH.write_text(json.dumps(TEST_JSON))

if not JSON_BZ2_PATH.is_file():
    with bz2.open(JSON_BZ2_PATH, mode="wt") as file:
        json.dump(TEST_JSON, file, indent=2)


class TestMocks(unittest.TestCase):
    """Tests for :mod:`pystow` mocks and context managers."""

    def test_mock_home(self):
        """Test that home can be properly mocked."""
        name = n()

        with tempfile.TemporaryDirectory() as d:
            expected_path = Path(d) / name
            self.assertFalse(expected_path.exists())

            with mock_envvar(PYSTOW_HOME_ENVVAR, expected_path.as_posix()):
                self.assertFalse(expected_path.exists())
                self.assertEqual(expected_path, get_home(ensure_exists=False))
                self.assertFalse(expected_path.exists())

    def test_mock_name(self):
        """Test that the name can be properly mocked."""
        name = n()

        expected_path = Path.home() / name
        self.assertFalse(expected_path.exists())

        with mock_envvar(PYSTOW_NAME_ENVVAR, name):
            self.assertEqual(name, get_name())

            self.assertFalse(expected_path.exists())
            self.assertEqual(expected_path, get_home(ensure_exists=False))
            self.assertFalse(expected_path.exists())


class TestGet(unittest.TestCase):
    """Tests for :mod:`pystow`."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.directory = tempfile.TemporaryDirectory()

    def tearDown(self) -> None:
        """Tear down the test case."""
        self.directory.cleanup()

    @contextlib.contextmanager
    def mock_directory(self) -> ContextManager[Path]:
        """Use this test case's temporary directory as a mock environment variable.

        :yield: The mock directory's path
        """
        with mock_envvar(PYSTOW_HOME_ENVVAR, self.directory.name):
            yield Path(self.directory.name)

    @staticmethod
    def mock_download():
        """Mock connection to the internet using local resource files.

        :return: A patch object that can be applied to the pystow download function
        """

        def _mock_get_data(url: str, path: Union[str, Path], **_kwargs) -> Path:
            return shutil.copy(MOCK_FILES[url], path)

        return mock.patch("pystow.utils.download", side_effect=_mock_get_data)

    @staticmethod
    def mock_download_once(local_path: Union[str, Path]):
        """Mock connection to the internet using local resource files.

        :param local_path: the path to the file to mock
        :return: A patch object that can be applied to the pystow download function
        """

        def _mock_get_data(path: Union[str, Path], **_kwargs) -> Path:
            return shutil.copy(local_path, path)

        return mock.patch("pystow.utils.download", side_effect=_mock_get_data)

    def join(self, *parts: str) -> Path:
        """Help join the parts to this test case's temporary directory.

        :param parts: The file path parts that are joined with this test case's directory
        :return: A path to the file
        """
        return Path(os.path.join(self.directory.name, *parts))

    def test_mock(self):
        """Test that mocking the directory works properly for this test case."""
        with self.mock_directory():
            self.assertEqual(os.getenv(PYSTOW_HOME_ENVVAR), self.directory.name)

    def test_get(self):
        """Test the :func:`get` function."""
        parts_examples = [
            [n()],
            [n(), n()],
            [n(), n(), n()],
        ]
        with self.mock_directory():
            for parts in parts_examples:
                with self.subTest(parts=parts):
                    self.assertEqual(self.join(*parts), join(*parts))

    def test_ensure(self):
        """Test ensuring various files."""
        write_pickle_gz(TEST_TSV_ROWS, path=PICKLE_GZ_PATH)

        with self.mock_directory(), self.mock_download():
            with self.subTest(type="tsv"):
                df = pystow.ensure_csv("test", url=TSV_URL)
                self.assertEqual(3, len(df.columns))

                df2 = pystow.load_df("test", name=TSV_NAME)
                self.assertEqual(df.values.tolist(), df2.values.tolist())

            with self.subTest(type="json"):
                j = pystow.ensure_json("test", url=JSON_URL)
                self.assertEqual(TEST_JSON, j)

                j2 = pystow.load_json("test", name=JSON_NAME)
                self.assertEqual(j, j2)

            with self.subTest(type="pickle"):
                p = pystow.ensure_pickle("test", url=PICKLE_URL)
                self.assertEqual(3, len(p))

                p2 = pystow.load_pickle("test", name=PICKLE_NAME)
                self.assertEqual(p, p2)

            with self.subTest(type="pickle_gz"):
                p = pystow.ensure_pickle_gz("test", url=PICKLE_GZ_URL)
                self.assertEqual(3, len(p))

                p2 = pystow.load_pickle_gz("test", name=PICKLE_GZ_NAME)
                self.assertEqual(p, p2)

            with self.subTest(type="json_bz2"):
                p = pystow.ensure_json_bz2("test", url=JSON_BZ2_URL)
                self.assertEqual(TEST_JSON, p)

    def test_open_fail(self):
        """Test opening a missing file."""
        with self.assertRaises(FileNotFoundError):
            with pystow.open("nope", name="nope"):
                pass

        with self.assertRaises(FileNotFoundError):
            pystow.load_json("nope", name="nope")

    def test_ensure_open_lzma(self):
        """Test opening lzma-encoded files."""
        with tempfile.TemporaryDirectory() as directory, self.mock_directory():
            path = Path(directory) / n()
            with self.mock_download_once(path):
                with lzma.open(path, "wt") as file:
                    for row in TEST_TSV_ROWS:
                        print(*row, sep="\t", file=file)  # noqa:T001,T201
                with pystow.ensure_open_lzma("test", url=n()) as file:
                    df = pd.read_csv(file, sep="\t")
                    self.assertEqual(3, len(df.columns))

    def test_ensure_open_zip(self):
        """Test opening tar-encoded files."""
        with tempfile.TemporaryDirectory() as directory, self.mock_directory():
            path = Path(directory) / n()
            inner_path = n()
            with self.mock_download_once(path):
                write_zipfile_csv(TEST_DF, path, inner_path)
                with pystow.ensure_open_zip("test", url=n(), inner_path=inner_path) as file:
                    df = pd.read_csv(file, sep="\t")
                    self.assertEqual(3, len(df.columns))

    def test_ensure_open_tarfile(self):
        """Test opening tarfile-encoded files."""
        with tempfile.TemporaryDirectory() as directory, self.mock_directory():
            path = Path(directory) / n()
            inner_path = n()
            with self.mock_download_once(path):
                write_tarfile_csv(TEST_DF, path, inner_path)
                with pystow.ensure_open_tarfile("test", url=n(), inner_path=inner_path) as file:
                    df = pd.read_csv(file, sep="\t")
                    self.assertEqual(3, len(df.columns))

    def test_ensure_module(self):
        """Test that the ``ensure_exist`` argument in :meth:`Module.from_key` works properly."""
        parts_examples = [
            [n()],
            [n(), n()],
            [n(), n(), n()],
        ]
        ensure_examples = [False, True]

        for ensure_exists, parts in itt.product(ensure_examples, parts_examples):
            with self.subTest(ensure_exists=ensure_exists, parts=parts), self.mock_directory():
                expected_directory = self.join(*parts)

                module = Module.from_key(*parts, ensure_exists=ensure_exists)

                self.assertEqual(expected_directory, module.base)
                self.assertIs(
                    expected_directory.exists(),
                    ensure_exists,
                    msg=f'{expected_directory} should{"" if ensure_exists else " not"} exist.',
                )

    def test_ensure_custom(self):
        """Test ensure with custom provider."""
        with self.mock_directory():
            # create a minimal provider
            def touch_file(path: Path, **_kwargs):
                """
                Create a file.

                :param path:
                    the file path
                :param _kwargs:
                    ignored keywords
                """
                path.touch()

            # wrap to record calls
            provider = mock.Mock(wraps=touch_file)

            # the keyword-based parameters for the provider
            kwargs = {"a": 4, "c": {0: 1, 5: 7}}

            # call first time
            name = n()
            path = pystow.ensure_custom("test", name=name, provider=provider, **kwargs)
            self.assertTrue(path.is_file())
            # call a second time
            path = pystow.ensure_custom("test", name=name, provider=provider, **kwargs)
            # ensure that the provider was only called once with the given parameters
            provider.assert_called_once_with(path, **kwargs)

    def test_ensure_open_sqlite(self):
        """Test caching SQLite."""
        with self.mock_directory(), self.mock_download():
            with pystow.ensure_open_sqlite("test", url=SQLITE_URL) as conn:
                df = pd.read_sql(f"SELECT * from {SQLITE_TABLE}", conn)  # noqa:S608
                self.assertEqual(3, len(df.columns))



# -*- coding: utf-8 -*-

"""Test configuration loading."""

import tempfile
import unittest
from pathlib import Path

import pystow
from pystow.config_api import CONFIG_HOME_ENVVAR, _get_cfp
from pystow.utils import mock_envvar


class TestConfig(unittest.TestCase):
    """Test configuration."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the class for testing."""
        cls.test_section = "test"
        cls.test_option = "option"
        cls.test_value = "value"
        cls.cfp = _get_cfp(cls.test_section)
        cls.cfp.add_section(cls.test_section)
        cls.cfp.set(
            section=cls.test_section,
            option=cls.test_option,
            value=cls.test_value,
        )

    def test_env_cast(self):
        """Test casting works properly when getting from the environment."""
        with mock_envvar("TEST_VAR", "1234"):
            self.assertEqual("1234", pystow.get_config("test", "var"))
            self.assertEqual("1234", pystow.get_config("test", "var", dtype=str))
            self.assertEqual(1234, pystow.get_config("test", "var", dtype=int))
            with self.assertRaises(ValueError):
                pystow.get_config("test", "var", dtype=bool)
            with self.assertRaises(TypeError):
                pystow.get_config("test", "var", dtype=object)

    def test_get_config(self):
        """Test lookup not existing."""
        self.assertIsNone(pystow.get_config(self.test_section, "key"))
        self.assertEqual("1234", pystow.get_config(self.test_section, "key", default="1234"))

        value = "not_value"
        self.assertEqual(
            value, pystow.get_config(self.test_section, self.test_option, passthrough=value)
        )

        self.assertEqual(1, pystow.get_config(self.test_section, self.test_option, passthrough=1))
        self.assertEqual(
            1, pystow.get_config(self.test_section, self.test_option, passthrough="1", dtype=int)
        )

        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="1", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="yes", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="Yes", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="YES", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="True", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="TRUE", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="T", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough="t", dtype=bool),
        )
        self.assertEqual(
            True,
            pystow.get_config(self.test_section, self.test_option, passthrough=True, dtype=bool),
        )
        self.assertEqual(
            True, pystow.get_config(self.test_section, self.test_option, passthrough=1, dtype=bool)
        )

    def test_subsection(self):
        """Test subsections."""
        with tempfile.TemporaryDirectory() as directory, mock_envvar(CONFIG_HOME_ENVVAR, directory):
            directory = Path(directory)
            path = directory.joinpath("test.ini")
            self.assertFalse(path.is_file(), msg="file should not already exist")

            self.assertIsNone(pystow.get_config("test:subtest", "key"))
            self.assertFalse(path.is_file(), msg="getting config should not create a file")

            pystow.write_config("test:subtest", "key", "value")
            self.assertTrue(path.is_file(), msg=f"{list(directory.iterdir())}")

            self.assertEqual("value", pystow.get_config("test:subtest", "key"))



# -*- coding: utf-8 -*-

"""Tests for utilities."""

import hashlib
import os
import tempfile
import unittest
from pathlib import Path

import numpy as np
import pandas as pd
import requests
from lxml import etree
from requests_file import FileAdapter

from pystow.utils import (
    HexDigestError,
    download,
    get_hexdigests_remote,
    getenv_path,
    mkdir,
    mock_envvar,
    n,
    name_from_url,
    read_tarfile_csv,
    read_zip_np,
    read_zipfile_csv,
    read_zipfile_xml,
    write_tarfile_csv,
    write_zipfile_csv,
    write_zipfile_np,
    write_zipfile_xml,
)

HERE = Path(__file__).resolve().parent
TEST_TXT = HERE.joinpath("resources", "test.txt")
TEST_TXT_MD5 = HERE.joinpath("resources", "test.txt.md5")
TEST_TXT_VERBOSE_MD5 = HERE.joinpath("resources", "test_verbose.txt.md5")
TEST_TXT_WRONG_MD5 = HERE.joinpath("resources", "test_wrong.txt.md5")

skip_on_windows = unittest.skipIf(
    os.name == "nt",
    reason="Funny stuff happens in requests with a file adapter on windows that adds line breaks",
)


class _Session(requests.sessions.Session):
    """A mock session."""

    def __init__(self):
        """Instantiate the patched session with an additional file adapter."""
        super().__init__()
        self.mount("file://", FileAdapter())


requests.sessions.Session = _Session


class TestUtils(unittest.TestCase):
    """Test utility functions."""

    def test_name_from_url(self):
        """Test :func:`name_from_url`."""
        data = [
            ("test.tsv", "https://example.com/test.tsv"),
            ("test.tsv", "https://example.com/deeper/test.tsv"),
            ("test.tsv.gz", "https://example.com/deeper/test.tsv.gz"),
        ]
        for name, url in data:
            with self.subTest(name=name, url=url):
                self.assertEqual(name, name_from_url(url))

    @skip_on_windows
    def test_file_values(self):
        """Test encodings."""
        for url, value in [
            (TEST_TXT, "this is a test file\n"),
            (TEST_TXT_MD5, "4221d002ceb5d3c9e9137e495ceaa647"),
            (TEST_TXT_VERBOSE_MD5, "MD5(text.txt)=4221d002ceb5d3c9e9137e495ceaa647"),
            (TEST_TXT_WRONG_MD5, "yolo"),
        ]:
            with self.subTest(name=url.name):
                self.assertEqual(value, requests.get(url.as_uri(), timeout=15).text)

    def test_mkdir(self):
        """Test for ensuring a directory."""
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            subdirectory = directory / "sd1"
            self.assertFalse(subdirectory.exists())

            mkdir(subdirectory, ensure_exists=False)
            self.assertFalse(subdirectory.exists())

            mkdir(subdirectory, ensure_exists=True)
            self.assertTrue(subdirectory.exists())

    def test_mock_envvar(self):
        """Test that environment variables can be mocked properly."""
        name, value = n(), n()

        self.assertNotIn(name, os.environ)
        with mock_envvar(name, value):
            self.assertIn(name, os.environ)
            self.assertEqual(value, os.getenv(name))
        self.assertNotIn(name, os.environ)

    def test_getenv_path(self):
        """Test that :func:`getenv_path` works properly."""
        envvar = n()

        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            value = directory / n()
            default = directory / n()

            self.assertEqual(default, getenv_path(envvar, default))
            with mock_envvar(envvar, value.as_posix()):
                self.assertEqual(value, getenv_path(envvar, default))
            # Check that it goes back
            self.assertEqual(default, getenv_path(envvar, default))

    def test_compressed_io(self):
        """Test that the read/write to compressed folder functions work."""
        rows = [[1, 2], [3, 4], [5, 6]]
        columns = ["A", "B"]
        df = pd.DataFrame(rows, columns=columns)
        inner_path = "okay.tsv"

        data = [
            ("test.zip", write_zipfile_csv, read_zipfile_csv),
            ("test.tar.gz", write_tarfile_csv, read_tarfile_csv),
        ]
        for name, writer, reader in data:
            with self.subTest(name=name), tempfile.TemporaryDirectory() as directory:
                directory = Path(directory)
                path = directory / name
                self.assertFalse(path.exists())
                writer(df, path=path, inner_path=inner_path)
                self.assertTrue(path.exists())
                new_df = reader(path=path, inner_path=inner_path)
                self.assertEqual(list(df.columns), list(new_df.columns))
                self.assertEqual(df.values.tolist(), new_df.values.tolist())

    def test_xml_io(self):
        """Test that read/write for XML element tree works."""
        root = etree.Element("Doc")
        level1 = etree.SubElement(root, "S")
        main = etree.SubElement(level1, "Text")
        main.text = "Thanks for contributing an answer to Stack Overflow!"
        second = etree.SubElement(level1, "Tokens")
        level2 = etree.SubElement(second, "Token", word="low")

        level3 = etree.SubElement(level2, "Morph")
        second1 = etree.SubElement(level3, "Lemma")
        second1.text = "sdfs"
        second1 = etree.SubElement(level3, "info")
        second1.text = "qw"

        level4 = etree.SubElement(level3, "Aff")
        second1 = etree.SubElement(level4, "Type")
        second1.text = "sdfs"
        second1 = etree.SubElement(level4, "Suf")
        second1.text = "qw"

        tree = etree.ElementTree(root)
        inner_path = "okay.tsv"
        data = [
            ("test.zip", write_zipfile_xml, read_zipfile_xml),
        ]
        for name, writer, reader in data:
            with self.subTest(name=name), tempfile.TemporaryDirectory() as directory:
                directory = Path(directory)
                path = directory / name
                self.assertFalse(path.exists())
                writer(tree, path=path, inner_path=inner_path)
                self.assertTrue(path.exists())
                new_tree = reader(path=path, inner_path=inner_path)
                self.assertEqual(
                    etree.tostring(tree, pretty_print=True),
                    etree.tostring(new_tree, pretty_print=True),
                )

    def test_numpy_io(self):
        """Test IO with numpy."""
        arr = np.array([[0, 1], [2, 3]])
        inner_path = "okay.npz"
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            path = directory / "test.zip"
            write_zipfile_np(arr, inner_path=inner_path, path=path)
            reloaded_arr = read_zip_np(path=path, inner_path=inner_path)
            self.assertTrue(np.array_equal(arr, reloaded_arr))


class TestHashing(unittest.TestCase):
    """Tests for hexdigest checking."""

    def setUp(self) -> None:
        """Set up a test."""
        self.directory = tempfile.TemporaryDirectory()
        self.path = Path(self.directory.name).joinpath("test.tsv")

        md5 = hashlib.md5()  # noqa:S303,S324
        with TEST_TXT.open("rb") as file:
            md5.update(file.read())
        self.expected_md5 = md5.hexdigest()
        self.mismatching_md5_hexdigest = "yolo"
        self.assertNotEqual(self.mismatching_md5_hexdigest, self.expected_md5)

    def tearDown(self) -> None:
        """Tear down a test."""
        self.directory.cleanup()

    def test_hash_success(self):
        """Test checking actually works."""
        self.assertFalse(self.path.exists())
        download(
            url=TEST_TXT.as_uri(),
            path=self.path,
            hexdigests={
                "md5": self.expected_md5,
            },
        )

    @skip_on_windows
    def test_hash_remote_success(self):
        """Test checking actually works."""
        self.assertFalse(self.path.exists())
        download(
            url=TEST_TXT.as_uri(),
            path=self.path,
            hexdigests_remote={
                "md5": TEST_TXT_MD5.as_uri(),
            },
            hexdigests_strict=True,
        )
        self.assertTrue(self.path.exists())

    @skip_on_windows
    def test_hash_remote_verbose_success(self):
        """Test checking actually works."""
        self.assertFalse(self.path.exists())
        download(
            url=TEST_TXT.as_uri(),
            path=self.path,
            hexdigests_remote={
                "md5": TEST_TXT_VERBOSE_MD5.as_uri(),
            },
            hexdigests_strict=False,
        )
        self.assertTrue(self.path.exists())

    def test_hash_remote_verbose_failure(self):
        """Test checking actually works."""
        self.assertFalse(self.path.exists())
        with self.assertRaises(HexDigestError):
            download(
                url=TEST_TXT.as_uri(),
                path=self.path,
                hexdigests_remote={
                    "md5": TEST_TXT_VERBOSE_MD5.as_uri(),
                },
                hexdigests_strict=True,
            )

    def test_hash_error(self):
        """Test hash error on download."""
        self.assertFalse(self.path.exists())
        with self.assertRaises(HexDigestError):
            download(
                url=TEST_TXT.as_uri(),
                path=self.path,
                hexdigests={
                    "md5": self.mismatching_md5_hexdigest,
                },
            )

    def test_hash_remote_error(self):
        """Test hash error on download."""
        self.assertFalse(self.path.exists())
        with self.assertRaises(HexDigestError):
            download(
                url=TEST_TXT.as_uri(),
                path=self.path,
                hexdigests_remote={
                    "md5": TEST_TXT_WRONG_MD5.as_uri(),
                },
                hexdigests_strict=True,
            )

    def test_override_hash_error(self):
        """Test hash error on download."""
        self.path.write_text("test file content")

        self.assertTrue(self.path.exists())
        with self.assertRaises(HexDigestError):
            download(
                url=TEST_TXT.as_uri(),
                path=self.path,
                hexdigests={
                    "md5": self.expected_md5,
                },
                force=False,
            )

    def test_override_hash_remote_error(self):
        """Test hash error on download."""
        self.path.write_text("test file content")

        self.assertTrue(self.path.exists())
        with self.assertRaises(HexDigestError):
            download(
                url=TEST_TXT.as_uri(),
                path=self.path,
                hexdigests_remote={
                    "md5": TEST_TXT_MD5.as_uri(),
                },
                hexdigests_strict=True,
                force=False,
            )

    def test_force(self):
        """Test overwriting wrong file."""
        # now if force=True it should not bother with the hash check
        self.path.write_text("test file content")

        self.assertTrue(self.path.exists())
        download(
            url=TEST_TXT.as_uri(),
            path=self.path,
            hexdigests={
                "md5": self.expected_md5,
            },
            force=True,
        )

    @skip_on_windows
    def test_remote_force(self):
        """Test overwriting wrong file."""
        # now if force=True it should not bother with the hash check
        self.path.write_text("test file content")

        self.assertTrue(self.path.exists())
        download(
            url=TEST_TXT.as_uri(),
            path=self.path,
            hexdigests_remote={
                "md5": TEST_TXT_MD5.as_uri(),
            },
            hexdigests_strict=True,
            force=True,
        )

    def test_hexdigest_urls(self):
        """Test getting hex digests from URLs."""
        for url, strict in [
            (TEST_TXT_MD5, True),
            (TEST_TXT_MD5, False),
            (TEST_TXT_VERBOSE_MD5, False),
        ]:
            hexdigests = get_hexdigests_remote(
                {"md5": url.as_uri()},
                hexdigests_strict=strict,
            )
            self.assertEqual(
                "4221d002ceb5d3c9e9137e495ceaa647",
                hexdigests["md5"],
            )

        hexdigests = get_hexdigests_remote(
            {"md5": TEST_TXT_VERBOSE_MD5.as_uri()}, hexdigests_strict=True
        )
        self.assertNotEqual(
            "4221d002ceb5d3c9e9137e495ceaa647",
            hexdigests["md5"],
        )

        # Live test case
        # hexdigests = get_hexdigests_remote(
        #     {"md5": "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed22n0001.xml.gz.md5"},
        #     hexdigests_strict=False,
        # )
        # self.assertEqual(
        #     {
        #         "md5": "0f08d8f3947dde1f3bced5e1f450c0da",
        #     },
        #     hexdigests,
        # )



# -*- coding: utf-8 -*-

"""Tests for caching."""

import os
import tempfile
import unittest
from pathlib import Path

from pystow.cache import CachedPickle

EXPECTED = 5
EXPECTED_2 = 6


class TestCache(unittest.TestCase):
    """Tests for caches."""

    def setUp(self) -> None:
        """Set up the test case with a temporary directory."""
        self.tmpdir = tempfile.TemporaryDirectory()
        self.directory = Path(self.tmpdir.name)

    def tearDown(self) -> None:
        """Tear down the test case's temporary directory."""
        self.tmpdir.cleanup()

    def test_cache_exception(self):
        """Test that exceptions aren't swallowed."""
        path = self.directory.joinpath("test.pkl")

        self.assertFalse(path.is_file())

        @CachedPickle(path=path)
        def _f1():
            raise NotImplementedError

        self.assertFalse(path.is_file(), msg="function has not been called")

        with self.assertRaises(NotImplementedError):
            _f1()

        self.assertFalse(
            path.is_file(),
            msg="file should not have been created if an exception was thrown by the function",
        )

    def test_cache_pickle(self):
        """Test caching a pickle."""
        path = self.directory.joinpath("test.pkl")
        self.assertFalse(
            path.is_file(),
            msg="the file should not exist at the beginning of the test",
        )

        raise_flag = True

        @CachedPickle(path=path)
        def _f1():
            if raise_flag:
                raise ValueError
            return EXPECTED

        self.assertFalse(path.is_file(), msg="the file should not exist until function is called")

        with self.assertRaises(ValueError):
            _f1()
        self.assertFalse(
            path.is_file(),
            msg="the function should throw an exception because of the flag, and no file should be created",
        )

        raise_flag = False
        actual = _f1()
        self.assertEqual(EXPECTED, actual)
        self.assertTrue(path.is_file(), msg="a file should have been created")

        raise_flag = True
        actual_2 = _f1()  # if raises, the caching mechanism didn't work
        self.assertEqual(EXPECTED, actual_2)
        self.assertTrue(path.is_file())

        os.unlink(path)
        self.assertFalse(path.is_file())
        with self.assertRaises(ValueError):
            _f1()

        @CachedPickle(path=path, force=True)
        def _f2():
            return EXPECTED_2

        self.assertEqual(EXPECTED_2, _f2())  # overwrites the file
        self.assertEqual(EXPECTED_2, _f1())



# -*- coding: utf-8 -*-

"""PyStow: Easily pick a place to store data for your python package."""

from .api import (  # noqa
    dump_df,
    dump_json,
    dump_pickle,
    dump_rdf,
    dump_xml,
    ensure,
    ensure_csv,
    ensure_custom,
    ensure_excel,
    ensure_from_google,
    ensure_from_s3,
    ensure_gunzip,
    ensure_json,
    ensure_json_bz2,
    ensure_open,
    ensure_open_bz2,
    ensure_open_gz,
    ensure_open_lzma,
    ensure_open_sqlite,
    ensure_open_sqlite_gz,
    ensure_open_tarfile,
    ensure_open_zip,
    ensure_pickle,
    ensure_pickle_gz,
    ensure_rdf,
    ensure_tar_df,
    ensure_tar_xml,
    ensure_untar,
    ensure_xml,
    ensure_zip_df,
    ensure_zip_np,
    join,
    joinpath_sqlite,
    load_df,
    load_json,
    load_pickle,
    load_pickle_gz,
    load_rdf,
    load_xml,
    module,
    open,
    open_gz,
)
from .config_api import ConfigError, get_config, write_config  # noqa
from .impl import Module  # noqa
from .utils import ensure_readme  # noqa

ensure_readme()



# -*- coding: utf-8 -*-

"""Test for API completeness."""

import inspect
import unittest

import pandas as pd
import rdflib
from lxml import etree

import pystow
from pystow import Module

SKIP = {"__init__"}


def _df_equal(a: pd.DataFrame, b: pd.DataFrame, msg=None) -> bool:
    return a.values.tolist() == b.values.tolist()


def _rdf_equal(a: rdflib.Graph, b: rdflib.Graph, msg=None) -> bool:
    return {tuple(t) for t in a} == {tuple(t) for t in b}


def _etree_equal(a: etree.ElementTree, b: etree.ElementTree, msg=None) -> bool:
    return etree.tostring(a) == etree.tostring(b)


class TestExposed(unittest.TestCase):
    """Test API exposure."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.addTypeEqualityFunc(pd.DataFrame, _df_equal)
        self.addTypeEqualityFunc(rdflib.Graph, _rdf_equal)
        self.addTypeEqualityFunc(type(etree.ElementTree()), _etree_equal)

    def assert_io(self, obj, ext: str, dump, load):
        """Test an object can be  dumped and loaded.

        :param obj: The object to dump
        :param ext: The extension to use
        :param dump: The dump function
        :param load: The load function
        """
        name = f"test.{ext}"
        path = pystow.join("test", name=name)
        if path.is_file():
            path.unlink()
        self.assertFalse(path.is_file())

        dump("test", name=name, obj=obj)
        self.assertTrue(path.is_file())
        self.assertEqual(obj, load("test", name=name))

    def test_exposed(self):
        """Test that all module-level functions also have a counterpart in the top-level API."""
        for name, func in Module.__dict__.items():
            if not inspect.isfunction(func) or name in SKIP:
                continue
            with self.subTest(name=name):
                self.assertIn(
                    name,
                    pystow.api.__all__,
                    msg=f"Module.{name} should be included in from `pystow.api.__all__`.",
                )
                self.assertTrue(
                    hasattr(pystow.api, name),
                    msg=f"`Module.{name} should be exposed as a top-level function in `pystow.api`.",
                )
                self.assertTrue(
                    hasattr(pystow, name),
                    msg=f"`pystow.api.{name}` should be imported in `pystow.__init__`.",
                )

    def test_io(self):
        """Test IO functions."""
        obj = ["a", "b", "c"]
        for ext, dump, load in [
            ("json", pystow.dump_json, pystow.load_json),
            ("pkl", pystow.dump_pickle, pystow.load_pickle),
        ]:
            with self.subTest(ext=ext):
                self.assert_io(obj, ext=ext, dump=dump, load=load)

    def test_pd_io(self):
        """Test pandas IO."""
        columns = list("abc")
        data = [(1, 2, 3), (4, 5, 6)]
        df = pd.DataFrame(data, columns=columns)
        self.assert_io(df, ext="tsv", load=pystow.load_df, dump=pystow.dump_df)

    def test_rdf_io(self):
        """Test RDFlib IO."""
        graph = rdflib.Graph()
        graph.add(
            (
                rdflib.URIRef("http://example.com/subject"),
                rdflib.URIRef("http://example.com/predicate"),
                rdflib.URIRef("http://example.com/object"),
            )
        )
        self.assertEqual(1, len(graph))
        self.assert_io(graph, ext="ttl", dump=pystow.dump_rdf, load=pystow.load_rdf)

    def test_xml_io(self):
        """Test XML I/O."""
        root = etree.Element("root")
        root.set("interesting", "somewhat")
        etree.SubElement(root, "test")
        my_tree = etree.ElementTree(root)
        self.assert_io(my_tree, ext="xml", dump=pystow.dump_xml, load=pystow.load_xml)



# -*- coding: utf-8 -*-

"""Utilities for caching files."""

import functools
import json
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    MutableMapping,
    Optional,
    TypeVar,
    Union,
    cast,
)

try:
    import pickle5 as pickle
except ImportError:
    import pickle

if TYPE_CHECKING:
    import pandas as pd

__all__ = [
    # Classses
    "Cached",
    "CachedPickle",
    "CachedJSON",
    "CachedCollection",
    "CachedDataFrame",
    # Types
    "Getter",
]

logger = logging.getLogger(__name__)

JSONType = Union[
    Dict[str, Any],
    List[Any],
]

X = TypeVar("X")
Getter = Callable[[], X]


class Cached(Generic[X], ABC):
    """Caching decorator."""

    def __init__(
        self,
        path: Union[str, Path],
        force: bool = False,
    ) -> None:
        """Instantiate the decorator.

        :param path: The path to the cache for the file
        :param force: Should a pre-existing file be disregared/overwritten?
        """
        self.path = Path(path)
        self.force = force

    def __call__(self, func: Getter[X]) -> Getter[X]:
        """Apply this instance as a decorator.

        :param func: The function to wrap
        :return: A wrapped function
        """

        @functools.wraps(func)
        def _wrapped() -> X:
            if self.path.is_file() and not self.force:
                return self.load()
            logger.debug("no cache found at %s", self.path)
            rv = func()
            logger.debug("writing cache to %s", self.path)
            self.dump(rv)
            return rv

        return _wrapped

    @abstractmethod
    def load(self) -> X:
        """Load data from the cache (typically by opening a file at the given path)."""

    @abstractmethod
    def dump(self, rv: X) -> None:
        """Dump data to the cache (typically by opening a file at the given path).

        :param rv: The data to dump
        """


class CachedJSON(Cached[JSONType]):
    """Make a function lazily cache its return value as JSON."""

    def load(self) -> JSONType:
        """Load data from the cache as JSON.

        :returns: A python object with JSON-like data from the cache
        """
        with open(self.path) as file:
            return cast(JSONType, json.load(file))

    def dump(self, rv: JSONType) -> None:
        """Dump data to the cache as JSON.

        :param rv: The JSON data to dump
        """
        with open(self.path, "w") as file:
            json.dump(rv, file, indent=2)


class CachedPickle(Cached[Any]):
    """Make a function lazily cache its return value as a pickle."""

    def load(self) -> Any:
        """Load data from the cache as a pickle.

        :returns: A python object loaded from the cache
        """
        with open(self.path, "rb") as file:
            return pickle.load(file)

    def dump(self, rv: Any) -> None:
        """Dump data to the cache as a pickle.

        :param rv: The arbitrary python object to dump
        """
        with open(self.path, "wb") as file:
            pickle.dump(rv, file, protocol=pickle.HIGHEST_PROTOCOL)


class CachedCollection(Cached[List[str]]):
    """Make a function lazily cache its return value as file."""

    def load(self) -> List[str]:
        """Load data from the cache as a list of strings.

        :returns: A list of strings loaded from the cache
        """
        with open(self.path) as file:
            return [line.strip() for line in file]

    def dump(self, rv: List[str]) -> None:
        """Dump data to the cache as a list of strings.

        :param rv: The list of strings to dump
        """
        with open(self.path, "w") as file:
            for line in rv:
                print(line, file=file)  # noqa:T001,T201


class CachedDataFrame(Cached["pd.DataFrame"]):
    """Make a function lazily cache its return value as a dataframe."""

    def __init__(
        self,
        path: Union[str, Path],
        force: bool = False,
        sep: Optional[str] = None,
        dtype: Optional[Any] = None,
        read_csv_kwargs: Optional[MutableMapping[str, Any]] = None,
    ) -> None:
        """Instantiate the decorator.

        :param path: The path to the cache for the file
        :param force: Should a pre-existing file be disregared/overwritten?
        :param sep: The separator. Defaults to TSV, since this is the only reasonable default.
        :param dtype: A shortcut for setting the dtype
        :param read_csv_kwargs: Additional kwargs to pass to :func:`pd.read_csv`.
        :raises ValueError: if sep is given as a kwarg and also in ``read_csv_kwargs``.
        """
        super().__init__(path=path, force=force)
        self.read_csv_kwargs = read_csv_kwargs or {}
        if "sep" not in self.read_csv_kwargs:
            self.sep = sep or "\t"
        elif sep is not None:
            raise ValueError
        else:
            self.sep = self.read_csv_kwargs.pop("sep")
        if dtype is not None:
            if "dtype" in self.read_csv_kwargs:
                raise ValueError
            self.read_csv_kwargs["dtype"] = dtype
        self.read_csv_kwargs.setdefault("keep_default_na", False)

    def load(self) -> "pd.DataFrame":
        """Load data from the cache as a dataframe.

        :returns: A dataframe loaded from the cache.
        """
        import pandas as pd

        return pd.read_csv(
            self.path,
            sep=self.sep,
            **self.read_csv_kwargs,
        )

    def dump(self, rv: "pd.DataFrame") -> None:
        """Dump data to the cache as a dataframe.

        :param rv: The dataframe to dump
        """
        rv.to_csv(self.path, sep=self.sep, index=False)



# -*- coding: utf-8 -*-

"""Module implementation."""

import bz2
import gzip
import json
import logging
import lzma
import sqlite3
import tarfile
import zipfile
from contextlib import closing, contextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Mapping,
    Optional,
    Sequence,
    Union,
)

from . import utils
from .constants import JSON, BytesOpener, Opener, Provider
from .utils import (
    base_from_gzip_name,
    download_from_google,
    download_from_s3,
    get_base,
    gunzip,
    mkdir,
    name_from_s3_key,
    name_from_url,
    path_to_sqlite,
    read_rdf,
    read_tarfile_csv,
    read_tarfile_xml,
    read_zip_np,
    read_zipfile_csv,
)

try:
    import pickle5 as pickle
except ImportError:
    import pickle

if TYPE_CHECKING:
    import botocore.client
    import lxml.etree
    import numpy
    import pandas as pd
    import rdflib

__all__ = ["Module"]

logger = logging.getLogger(__name__)


class Module:
    """The class wrapping the directory lookup implementation."""

    def __init__(self, base: Union[str, Path], ensure_exists: bool = True) -> None:
        """Initialize the module.

        :param base:
            The base directory for the module
        :param ensure_exists:
            Should the base directory be created automatically?
            Defaults to true.
        """
        self.base = Path(base)
        mkdir(self.base, ensure_exists=ensure_exists)

    @classmethod
    def from_key(cls, key: str, *subkeys: str, ensure_exists: bool = True) -> "Module":
        """Get a module for the given directory or one of its subdirectories.

        :param key:
            The name of the module. No funny characters. The envvar
            <key>_HOME where key is uppercased is checked first before using
            the default home directory.
        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param ensure_exists:
            Should all directories be created automatically?
            Defaults to true.
        :return:
            A module
        """
        base = get_base(key, ensure_exists=False)
        rv = cls(base=base, ensure_exists=ensure_exists)
        if subkeys:
            rv = rv.module(*subkeys, ensure_exists=ensure_exists)
        return rv

    def module(self, *subkeys: str, ensure_exists: bool = True) -> "Module":
        """Get a module for a subdirectory of the current module.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param ensure_exists:
            Should all directories be created automatically?
            Defaults to true.
        :return:
            A module representing the subdirectory based on the given ``subkeys``.
        """
        base = self.join(*subkeys, ensure_exists=False)
        return Module(base=base, ensure_exists=ensure_exists)

    def join(
        self,
        *subkeys: str,
        name: Optional[str] = None,
        ensure_exists: bool = True,
    ) -> Path:
        """Get a subdirectory of the current module.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param ensure_exists:
            Should all directories be created automatically?
            Defaults to true.
        :param name:
            The name of the file (optional) inside the folder
        :return:
            The path of the directory or subdirectory for the given module.
        """
        rv = self.base
        if subkeys:
            rv = rv.joinpath(*subkeys)
            mkdir(rv, ensure_exists=ensure_exists)
        if name:
            rv = rv.joinpath(name)
        return rv

    def joinpath_sqlite(self, *subkeys: str, name: str) -> str:
        """Get an SQLite database connection string.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the database file.
        :return: A SQLite path string.
        """
        path = self.join(*subkeys, name=name, ensure_exists=True)
        return path_to_sqlite(path)

    def ensure(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Path:
        """Ensure a file is downloaded.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :return:
            The path of the file that has been downloaded (or already exists)
        """
        if name is None:
            name = name_from_url(url)
        path = self.join(*subkeys, name=name, ensure_exists=True)
        utils.download(
            url=url,
            path=path,
            force=force,
            **(download_kwargs or {}),
        )
        return path

    def ensure_custom(
        self,
        *subkeys: str,
        name: str,
        force: bool = False,
        provider: Provider,
        **kwargs: Any,
    ) -> Path:
        """Ensure a file is present, and run a custom create function otherwise.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name:
            The file name.
        :param force:
            Should the file be re-created, even if the path already exists?
        :param provider:
            The file provider. Will be run with the path as the first positional argument,
            if the file needs to be generated.
        :param kwargs:
            Additional keyword-based parameters passed to the provider.

        :raises ValueError:
            If the provider was called but the file was not created by it.

        :return:
            The path of the file that has been created (or already exists)
        """
        path = self.join(*subkeys, name=name, ensure_exists=True)
        if path.is_file() and not force:
            return path
        provider(path, **kwargs)
        if not path.is_file():
            raise ValueError(f"Provider {provider} did not create the file at {path}!")
        return path

    def ensure_untar(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        directory: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        extract_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Path:
        """Ensure a tar file is downloaded and unarchived.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param directory:
            Overrides the name of the directory into which the tar archive is extracted.
            If none given, will use the stem of the file name that gets downloaded.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param extract_kwargs: Keyword arguments to pass to :meth:`tarfile.TarFile.extract_all`.
        :return:
            The path of the directory where the file that has been downloaded
            gets extracted to
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        if directory is None:
            # rhea-rxn.tar.gz -> rhea-rxn
            suffixes_len = sum(len(suffix) for suffix in path.suffixes)
            directory = path.name[:-suffixes_len]
        unzipped_path = path.parent.joinpath(directory)
        if unzipped_path.is_dir() and not force:
            return unzipped_path
        unzipped_path.mkdir(exist_ok=True, parents=True)
        with tarfile.open(path) as tar_file:
            tar_file.extractall(unzipped_path, **(extract_kwargs or {}))  # noqa:S202
        return unzipped_path

    def ensure_gunzip(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        autoclean: bool = True,
        download_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Path:
        """Ensure a tar.gz file is downloaded and unarchived.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param autoclean: Should the zipped file be deleted?
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :return:
            The path of the directory where the file that has been downloaded
            gets extracted to
        """
        if name is None:
            name = name_from_url(url)
        gunzipped_name = base_from_gzip_name(name)
        gunzipped_path = self.join(*subkeys, name=gunzipped_name, ensure_exists=True)
        if gunzipped_path.is_file() and not force:
            return gunzipped_path
        path = self.ensure(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
        )
        gunzip(path, gunzipped_path)
        if autoclean:
            logger.info("removing original gzipped file %s", path)
            path.unlink()
        return gunzipped_path

    @contextmanager
    def ensure_open(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "r",
        open_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Opener:
        """Ensure a file is downloaded and open it.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with path.open(**open_kwargs) as file:
            yield file

    @contextmanager
    def open(
        self,
        *subkeys: str,
        name: str,
        mode: str = "r",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        ensure_exists: bool = False,
    ) -> Opener:
        """Open a file that exists already.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param ensure_exists: Should the file be made? Set to true on write operations.

        :yields: An open file object
        """
        path = self.join(*subkeys, name=name, ensure_exists=ensure_exists)
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with path.open(**open_kwargs) as file:
            yield file

    @contextmanager
    def open_gz(
        self,
        *subkeys: str,
        name: str,
        mode: str = "rt",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        ensure_exists: bool = False,
    ) -> Opener:
        """Open a gzipped file that exists already.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`gzip.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
        :param ensure_exists: Should the file be made? Set to true on write operations.

        :yields: An open file object
        """
        path = self.join(*subkeys, name=name, ensure_exists=ensure_exists)
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with gzip.open(path, **open_kwargs) as file:
            yield file

    @contextmanager
    def ensure_open_lzma(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "rt",
        open_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Opener:
        """Ensure a LZMA-compressed file is downloaded and open a file inside it.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`lzma.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`lzma.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with lzma.open(path, **open_kwargs) as file:
            yield file

    @contextmanager
    def ensure_open_tarfile(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "r",
        open_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> BytesOpener:
        """Ensure a tar file is downloaded and open a file inside it.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param inner_path:
            The relative path to the file inside the archive
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`tarfile.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`tarfile.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with tarfile.open(path, **open_kwargs) as tar_file:
            with tar_file.extractfile(inner_path) as file:  # type:ignore
                yield file

    @contextmanager
    def ensure_open_zip(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "r",
        open_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> BytesOpener:
        """Ensure a file is downloaded then open it with :mod:`zipfile`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param inner_path:
            The relative path to the file inside the archive
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`zipfile.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`zipfile.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with zipfile.ZipFile(file=path) as zip_file:
            with zip_file.open(inner_path) as file:
                yield file

    @contextmanager
    def ensure_open_gz(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "rb",
        open_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Opener:
        """Ensure a gzipped file is downloaded and open a file inside it.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`gzip.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with gzip.open(path, **open_kwargs) as file:
            yield file

    @contextmanager
    def ensure_open_bz2(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "rb",
        open_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Opener:
        """Ensure a BZ2-compressed file is downloaded and open a file inside it.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`bz2.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`bz2.open`

        :yields: An open file object
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
        open_kwargs.setdefault("mode", mode)
        with bz2.open(path, **open_kwargs) as file:
            yield file

    def ensure_csv(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        read_csv_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "pd.DataFrame":
        """Download a CSV and open as a dataframe with :mod:`pandas`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
        :return: A pandas DataFrame
        :rtype: pandas.DataFrame
        """
        import pandas as pd

        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return pd.read_csv(path, **_clean_csv_kwargs(read_csv_kwargs))

    def load_df(
        self,
        *subkeys: str,
        name: str,
        read_csv_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "pd.DataFrame":
        """Open a pre-existing CSV as a dataframe with :mod:`pandas`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
        :return: A pandas DataFrame
        """
        import pandas as pd

        with self.open(*subkeys, name=name) as file:
            return pd.read_csv(file, **_clean_csv_kwargs(read_csv_kwargs))

    def dump_df(
        self,
        *subkeys: str,
        name: str,
        obj: "pd.DataFrame",
        sep: str = "\t",
        index: bool = False,
        to_csv_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Dump a dataframe to a TSV file with :mod:`pandas`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param obj: The dataframe to dump
        :param sep: The separator to use, defaults to a tab
        :param index: Should the index be dumped? Defaults to false.
        :param to_csv_kwargs: Keyword arguments to pass through to :meth:`pandas.DataFrame.to_csv`.
        """
        to_csv_kwargs = {} if to_csv_kwargs is None else dict(to_csv_kwargs)
        to_csv_kwargs.setdefault("sep", sep)
        to_csv_kwargs.setdefault("index", index)
        # should this use unified opener instead? Pandas is pretty smart...
        path = self.join(*subkeys, name=name)
        obj.to_csv(path, **to_csv_kwargs)

    def ensure_json(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        open_kwargs: Optional[Mapping[str, Any]] = None,
        json_load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> JSON:
        """Download JSON and open with :mod:`json`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.
        :returns: A JSON object (list, dict, etc.)
        """
        with self.ensure_open(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            open_kwargs=open_kwargs,
        ) as file:
            return json.load(file, **(json_load_kwargs or {}))

    def ensure_json_bz2(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        open_kwargs: Optional[Mapping[str, Any]] = None,
        json_load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> JSON:
        """Download BZ2-compressed JSON and open with :mod:`json`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param open_kwargs: Additional keyword arguments passed to :func:`bz2.open`
        :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.
        :returns: A JSON object (list, dict, etc.)
        """
        with self.ensure_open_bz2(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            open_kwargs=open_kwargs,
        ) as file:
            return json.load(file, **(json_load_kwargs or {}))

    def load_json(
        self,
        *subkeys: str,
        name: str,
        open_kwargs: Optional[Mapping[str, Any]] = None,
        json_load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> JSON:
        """Open a JSON file :mod:`json`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param json_load_kwargs: Keyword arguments to pass through to :func:`json.load`.
        :returns: A JSON object (list, dict, etc.)
        """
        with self.open(
            *subkeys, name=name, mode="r", open_kwargs=open_kwargs, ensure_exists=True
        ) as file:
            return json.load(file, **(json_load_kwargs or {}))

    def dump_json(
        self,
        *subkeys: str,
        name: str,
        obj: JSON,
        open_kwargs: Optional[Mapping[str, Any]] = None,
        json_dump_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Dump an object to a file with :mod:`json`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param json_dump_kwargs: Keyword arguments to pass through to :func:`json.dump`.
        """
        with self.open(
            *subkeys, name=name, mode="w", open_kwargs=open_kwargs, ensure_exists=True
        ) as file:
            json.dump(obj, file, **(json_dump_kwargs or {}))

    def ensure_pickle(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "rb",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """Download a pickle file and open with :mod:`pickle`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
        :returns: Any object
        """
        with self.ensure_open(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def load_pickle(
        self,
        *subkeys: str,
        name: str,
        mode: str = "rb",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """Open a pickle file with :mod:`pickle`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
        :returns: Any object
        """
        with self.open(
            *subkeys,
            name=name,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def dump_pickle(
        self,
        *subkeys: str,
        name: str,
        obj: Any,
        mode: str = "wb",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        pickle_dump_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Dump an object to a file with :mod:`pickle`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param pickle_dump_kwargs: Keyword arguments to pass through to :func:`pickle.dump`.
        """
        with self.open(
            *subkeys,
            name=name,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            pickle.dump(obj, file, **(pickle_dump_kwargs or {}))

    def ensure_pickle_gz(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        mode: str = "rb",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """Download a gzipped pickle file and open with :mod:`pickle`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param mode: The read mode, passed to :func:`gzip.open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
        :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
        :returns: Any object
        """
        with self.ensure_open_gz(
            *subkeys,
            url=url,
            name=name,
            force=force,
            download_kwargs=download_kwargs,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def load_pickle_gz(
        self,
        *subkeys: str,
        name: str,
        mode: str = "rb",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        pickle_load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """Open a gzipped pickle file with :mod:`pickle`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param mode: The read mode, passed to :func:`open`
        :param open_kwargs: Additional keyword arguments passed to :func:`gzip.open`
        :param pickle_load_kwargs: Keyword arguments to pass through to :func:`pickle.load`.
        :returns: Any object
        """
        with self.open_gz(
            *subkeys,
            name=name,
            mode=mode,
            open_kwargs=open_kwargs,
        ) as file:
            return pickle.load(file, **(pickle_load_kwargs or {}))

    def ensure_excel(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        read_excel_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "pd.DataFrame":
        """Download an excel file and open as a dataframe with :mod:`pandas`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param read_excel_kwargs: Keyword arguments to pass through to :func:`pandas.read_excel`.
        :return: A pandas DataFrame
        """
        import pandas as pd

        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return pd.read_excel(path, **(read_excel_kwargs or {}))

    def ensure_tar_df(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        read_csv_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "pd.DataFrame":
        """Download a tar file and open an inner file as a dataframe with :mod:`pandas`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param inner_path:
            The relative path to the file inside the archive
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
        :returns: A dataframe

        .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_tarfile_csv(
            path=path, inner_path=inner_path, **_clean_csv_kwargs(read_csv_kwargs)
        )

    def ensure_xml(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        parse_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "lxml.etree.ElementTree":
        """Download an XML file and open it with :mod:`lxml`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param parse_kwargs: Keyword arguments to pass through to :func:`lxml.etree.parse`.
        :returns: An ElementTree object

        .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
        """
        from lxml import etree

        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return etree.parse(path, **(parse_kwargs or {}))

    def load_xml(
        self,
        *subkeys: str,
        name: str,
        parse_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "lxml.etree.ElementTree":
        """Load an XML file with :mod:`lxml`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param parse_kwargs: Keyword arguments to pass through to :func:`lxml.etree.parse`.
        :returns: An ElementTree object

        .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
        """
        from lxml import etree

        with self.open(*subkeys, name=name, ensure_exists=False) as file:
            return etree.parse(file, **(parse_kwargs or {}))

    def dump_xml(
        self,
        *subkeys: str,
        name: str,
        obj: "lxml.etree.ElementTree",
        open_kwargs: Optional[Mapping[str, Any]] = None,
        write_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Dump an XML element tree to a file with :mod:`lxml`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param open_kwargs: Additional keyword arguments passed to :func:`open`
        :param write_kwargs: Keyword arguments to pass through to :func:`lxml.etree.ElementTree.write`.
        """
        with self.open(
            *subkeys, name=name, mode="wb", open_kwargs=open_kwargs, ensure_exists=True
        ) as file:
            obj.write(file, **(write_kwargs or {}))

    def ensure_tar_xml(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        parse_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "lxml.etree.ElementTree":
        """Download a tar file and open an inner file as an XML with :mod:`lxml`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param inner_path:
            The relative path to the file inside the archive
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param parse_kwargs: Keyword arguments to pass through to :func:`lxml.etree.parse`.
        :returns: An ElementTree object

        .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_tarfile_xml(path=path, inner_path=inner_path, **(parse_kwargs or {}))

    def ensure_zip_df(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        read_csv_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "pd.DataFrame":
        """Download a zip file and open an inner file as a dataframe with :mod:`pandas`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param inner_path:
            The relative path to the file inside the archive
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
        :return: A pandas DataFrame
        :rtype: pandas.DataFrame
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_zipfile_csv(
            path=path, inner_path=inner_path, **_clean_csv_kwargs(read_csv_kwargs)
        )

    def ensure_zip_np(
        self,
        *subkeys: str,
        url: str,
        inner_path: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        load_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "numpy.typing.ArrayLike":
        """Download a zip file and open an inner file as an array-like with :mod:`numpy`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param inner_path:
            The relative path to the file inside the archive
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs:
            Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param load_kwargs:
            Additional keyword arguments that are passed through to :func:`read_zip_np`
            and transitively to :func:`numpy.load`.
        :returns: An array-like object
        :rtype: numpy.typing.ArrayLike
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        return read_zip_np(path=path, inner_path=inner_path, **(load_kwargs or {}))

    def ensure_rdf(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
        precache: bool = True,
        parse_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "rdflib.Graph":
        """Download a RDF file and open with :mod:`rdflib`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :param precache: Should the parsed :class:`rdflib.Graph` be stored as a pickle for fast loading?
        :param parse_kwargs:
            Keyword arguments to pass through to :func:`pystow.utils.read_rdf` and transitively to
            :func:`rdflib.Graph.parse`.
        :return: An RDF graph
        :rtype: rdflib.Graph
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        if not precache:
            return read_rdf(path=path, **(parse_kwargs or {}))

        cache_path = path.with_suffix(path.suffix + ".pickle.gz")
        if cache_path.exists() and not force:
            with gzip.open(cache_path, "rb") as file:
                return pickle.load(file)

        rv = read_rdf(path=path, **(parse_kwargs or {}))
        with gzip.open(cache_path, "wb") as file:
            pickle.dump(rv, file, protocol=pickle.HIGHEST_PROTOCOL)
        return rv

    def load_rdf(
        self,
        *subkeys: str,
        name: Optional[str] = None,
        parse_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> "rdflib.Graph":
        """Open an RDF file with :mod:`rdflib`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param parse_kwargs:
            Keyword arguments to pass through to :func:`pystow.utils.read_rdf` and transitively to
            :func:`rdflib.Graph.parse`.
        :return: An RDF graph
        """
        path = self.join(*subkeys, name=name, ensure_exists=False)
        return read_rdf(path=path, **(parse_kwargs or {}))

    def dump_rdf(
        self,
        *subkeys: str,
        name: str,
        obj: "rdflib.Graph",
        format: str = "turtle",
        serialize_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Dump an RDF graph to a file with :mod:`rdflib`.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name: The name of the file to open
        :param obj: The object to dump
        :param format: The format to dump in
        :param serialize_kwargs:
            Keyword arguments to through to :func:`rdflib.Graph.serialize`.
        """
        path = self.join(*subkeys, name=name, ensure_exists=False)
        serialize_kwargs = {} if serialize_kwargs is None else dict(serialize_kwargs)
        serialize_kwargs.setdefault("format", format)
        obj.serialize(path, **serialize_kwargs)

    def ensure_from_s3(
        self,
        *subkeys: str,
        s3_bucket: str,
        s3_key: Union[str, Sequence[str]],
        name: Optional[str] = None,
        client: Optional["botocore.client.BaseClient"] = None,
        client_kwargs: Optional[Mapping[str, Any]] = None,
        download_file_kwargs: Optional[Mapping[str, Any]] = None,
        force: bool = False,
    ) -> Path:
        """Ensure a file is downloaded.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param s3_bucket:
            The S3 bucket name
        :param s3_key:
            The S3 key name
        :param name:
            Overrides the name of the file at the end of the S3 key, if given.
        :param client:
            A botocore client. If none given, one will be created automatically
        :param client_kwargs:
            Keyword arguments to be passed to the client on instantiation.
        :param download_file_kwargs:
            Keyword arguments to be passed to :func:`boto3.s3.transfer.S3Transfer.download_file`
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :return:
            The path of the file that has been downloaded (or already exists)
        """
        if not isinstance(s3_key, str):
            s3_key = "/".join(s3_key)  # join sequence
        if name is None:
            name = name_from_s3_key(s3_key)
        path = self.join(*subkeys, name=name, ensure_exists=True)
        download_from_s3(
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            path=path,
            client=client,
            client_kwargs=client_kwargs,
            force=force,
            download_file_kwargs=download_file_kwargs,
        )
        return path

    def ensure_from_google(
        self,
        *subkeys: str,
        name: str,
        file_id: str,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Path:
        """Ensure a file is downloaded from Google Drive.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param name:
            The name of the file
        :param file_id:
            The file identifier of the google file. If your share link is
            https://drive.google.com/file/d/1AsPPU4ka1Rc9u-XYMGWtvV65hF3egi0z/view, then your file id is
            ``1AsPPU4ka1Rc9u-XYMGWtvV65hF3egi0z``.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs:
            Keyword arguments to pass through to :func:`pystow.utils.download_from_google`.
        :return:
            The path of the file that has been downloaded (or already exists)
        """
        path = self.join(*subkeys, name=name, ensure_exists=True)
        download_from_google(file_id, path, force=force, **(download_kwargs or {}))
        return path

    @contextmanager
    def ensure_open_sqlite(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Generator[sqlite3.Connection, None, None]:
        """Ensure and connect to a SQLite database.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :yields: An instance of :class:`sqlite3.Connection` from :func:`sqlite3.connect`

        Example usage:
        >>> import pystow
        >>> import pandas as pd
        >>> url = "https://s3.amazonaws.com/bbop-sqlite/hp.db"
        >>> sql = "SELECT * FROM entailed_edge LIMIT 10"
        >>> module = pystow.module("test")
        >>> with module.ensure_open_sqlite(url=url) as conn:
        >>>     df = pd.read_sql(sql, conn)
        """
        path = self.ensure(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        with closing(sqlite3.connect(path.as_posix())) as conn:
            yield conn

    @contextmanager
    def ensure_open_sqlite_gz(
        self,
        *subkeys: str,
        url: str,
        name: Optional[str] = None,
        force: bool = False,
        download_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Generator[sqlite3.Connection, None, None]:
        """Ensure and connect to a SQLite database that's gzipped.

        Unfortunately, it's a paid feature to directly read gzipped sqlite files,
        so this automatically gunzips it first.

        :param subkeys:
            A sequence of additional strings to join. If none are given,
            returns the directory for this module.
        :param url:
            The URL to download.
        :param name:
            Overrides the name of the file at the end of the URL, if given. Also
            useful for URLs that don't have proper filenames with extensions.
        :param force:
            Should the download be done again, even if the path already exists?
            Defaults to false.
        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
        :yields: An instance of :class:`sqlite3.Connection` from :func:`sqlite3.connect`

        Example usage:
        >>> import pystow
        >>> import pandas as pd
        >>> url = "https://s3.amazonaws.com/bbop-sqlite/hp.db.gz"
        >>> module = pystow.module("test")
        >>> sql = "SELECT * FROM entailed_edge LIMIT 10"
        >>> with module.ensure_open_sqlite_gz(url=url) as conn:
        >>>     df = pd.read_sql(sql, conn)
        """
        path = self.ensure_gunzip(
            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
        )
        with closing(sqlite3.connect(path.as_posix())) as conn:
            yield conn


def _clean_csv_kwargs(read_csv_kwargs: Union[None, Mapping[str, Any]]) -> Dict[str, Any]:
    read_csv_kwargs = {} if read_csv_kwargs is None else dict(read_csv_kwargs)
    read_csv_kwargs.setdefault("sep", "\t")
    return read_csv_kwargs
